''' This set of methods supports generating and querying healpix MOC maps for GW events,
    using healpix_alchemy and model mappings with sql_alchemy queries.
'''
from astropy.table import Table
from tom_nonlocalizedevents.models import (NonLocalizedEvent, EventCandidate, SkymapTile,
                                           EventLocalization, CredibleRegion)
from django.db import transaction
from django.conf import settings
from django.utils import timezone
from healpix_alchemy.constants import HPX, LEVEL
from healpix_alchemy.types import Tile, Point
import sqlalchemy as sa
from sqlalchemy.orm import relationship, declarative_base, Session
from astropy import units as u
import astropy_healpix as ah
from mocpy import MOC
from ligo.skymap import distance
from dateutil.parser import parse
import numpy as np
import os
import hashlib
from io import BytesIO
import uuid
import sys
import json
import logging

# from django.db.models import Sum, Subquery, F, Min
# from tom_nonlocalizedevents_base.settings import DATABASES


logger = logging.getLogger(__name__)

#
# SQLAlchemy Engine Configuration
#  see https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.create_engine

SA_DB_CONNECTION_URL = os.getenv(
    'SA_DB_CONNECTION_URL',
    (f"postgresql://{settings.DATABASES['default']['USER']}:{settings.DATABASES['default']['PASSWORD']}"
     f"@{settings.DATABASES['default']['HOST']}:{settings.DATABASES['default']['PORT']}"
     f"/{settings.DATABASES['default']['NAME']}"))
POOL_RECYCLE = 4 * 60 * 60
POOL_SIZE = os.getenv('POOL_SIZE', 5)
MAX_OVERFLOW = os.getenv('MAX_OVERFLOW', 10)

CREDIBLE_REGION_PROBABILITIES = sorted(json.loads(os.getenv(
    'CREDIBLE_REGION_PROBABILITIES', '[0.25, 0.5, 0.75, 0.9, 0.95]')), reverse=True)

Base = declarative_base()
sa_engine = sa.create_engine(
    SA_DB_CONNECTION_URL,
    pool_recycle=POOL_RECYCLE,
    pool_size=POOL_SIZE,
    max_overflow=MAX_OVERFLOW
)


def uniq_to_bigintrange(value):
    level, ipix = ah.uniq_to_level_ipix(value)
    shift = 2 * (LEVEL - level)
    return (ipix << shift, (ipix + 1) << shift)


def sequence_to_bigintrange(sequence):
    return f'[{sequence[0]},{sequence[1]})'


def tiles_from_moc(moc):
    return (f'[{lo},{hi})' for lo, hi in moc.to_depth29_ranges)


def tiles_from_polygon_skycoord(polygon):
    return tiles_from_moc(MOC.from_polygon_skycoord(polygon.transform_to(HPX.frame)))


def get_confidence_regions(skymap: Table):
    """ This helper method takes in the astropy Table skymap and attempts to parse out
        the 50 and 90 area confidence values. It returns a tuple of (area_50, area_90).
        See https://emfollow.docs.ligo.org/userguide/tutorial/multiorder_skymaps.html.
    """
    try:
        # Sort the pixels of the sky map by descending probability density
        skymap.sort('PROBDENSITY', reverse=True)
        # Find the area of each pixel
        level, ipix = ah.uniq_to_level_ipix(skymap['UNIQ'])
        pixel_area = ah.nside_to_pixel_area(ah.level_to_nside(level))
        # Calculate the probability within each pixel: the pixel area times the probability density
        prob = pixel_area * skymap['PROBDENSITY']
        # Calculate the cumulative sum of the probability
        cumprob = np.cumsum(prob)
        # Find the pixel for which the probability sums to 0.5 (0.9)
        index_50 = cumprob.searchsorted(0.5)
        index_90 = cumprob.searchsorted(0.9)
        # The area of the 50% (90%) credible region is simply the sum of the areas of the pixels up to that probability
        area_50 = pixel_area[:index_50].sum().to_value(u.deg ** 2.)
        area_90 = pixel_area[:index_90].sum().to_value(u.deg ** 2.)

        return area_50, area_90
    except Exception as e:
        logger.error(f'Unable to parse raw skymap for OBJECT {skymap.meta["OBJECT"]} for confidence regions: {e}')

    return None, None


def get_skymap_version(nle: NonLocalizedEvent, skymap_hash: uuid, is_combined: bool) -> int:
    """ This method gets the most recent previous sequence of this superevent and checks if the skymap has changed.
        It returns the 'version' of the skymap, which can be used to retrieve the proper file and image files from
        gracedb. This is a hack because IGWN GWAlerts no longer have any way of knowing which skymap version they
        reference.
    """
    try:
        for sequence in nle.sequences.all().reverse():
            if (is_combined and sequence.external_coincidence and sequence.external_coincidence.localization
                    and sequence.external_coincidence.localization.skymap_hash != skymap_hash):
                return sequence.external_coincidence.localization.skymap_version + 1
            elif (not is_combined) and sequence.localization and sequence.localization.skymap_hash != skymap_hash:
                return sequence.localization.skymap_version + 1
        return 0
    except NonLocalizedEvent.DoesNotExist:
        return 0  # The nonlocalizedevent doesnt exist in our system yet, so this must be the first skymap version


def create_localization_for_skymap(nonlocalizedevent: NonLocalizedEvent, skymap_bytes: bytes, skymap_url: str = '',
                                   is_combined=False):
    """ Create localization from skymap bytes and related fields """
    logger.info(f"Creating localization for {nonlocalizedevent.event_id} with skymap {skymap_url}")
    skymap_hash = hashlib.md5(skymap_bytes).hexdigest()
    try:
        localization = EventLocalization.objects.get(nonlocalizedevent=nonlocalizedevent, skymap_hash=skymap_hash)
    except EventLocalization.DoesNotExist:
        skymap = Table.read(BytesIO(skymap_bytes))
        distance_mean = skymap.meta['DISTMEAN']
        distance_std = skymap.meta['DISTSTD']
        date = parse(skymap.meta['DATE']).replace(tzinfo=timezone.utc)
        skymap_uuid = uuid.UUID(skymap_hash)
        skymap_version = get_skymap_version(nonlocalizedevent, skymap_hash=skymap_uuid, is_combined=is_combined)
        if not skymap_url:
            if is_combined:
                skymap_url = (
                    f"https://gracedb.ligo.org/api/superevents/{nonlocalizedevent.event_id}"
                    f"/files/combined-ext.multiorder.fits,{skymap_version}"
                )
            else:
                skymap_url = (
                    f"https://gracedb.ligo.org/api/superevents/{nonlocalizedevent.event_id}"
                    f"/files/bayestar.multiorder.fits,{skymap_version}"
                )
        area_50, area_90 = get_confidence_regions(skymap)
        row_dist_mean, row_dist_std, _ = distance.parameters_to_moments(
            skymap['DISTMU'], skymap['DISTSIGMA'])
        with transaction.atomic():
            localization = EventLocalization.objects.create(
                nonlocalizedevent=nonlocalizedevent,
                distance_mean=distance_mean,
                distance_std=distance_std,
                skymap_version=skymap_version,
                skymap_hash=skymap_uuid,
                skymap_url=skymap_url,
                area_50=area_50,
                area_90=area_90,
                date=date
            )
            for i, row in enumerate(skymap):
                # This is necessary to make sure we don't get an underflow error in postgres
                # when operating with the probdensity float field
                probdensity = row['PROBDENSITY'] if row['PROBDENSITY'] > sys.float_info.min else 0
                SkymapTile.objects.create(
                    localization=localization,
                    tile=uniq_to_bigintrange(row['UNIQ']),
                    probdensity=probdensity,
                    distance_mean=row_dist_mean[i],
                    distance_std=row_dist_std[i]
                )
    return localization


# healpix_alchemy models pointing to django ORM models, for building a sql alchemy query
class SaSkymap(Base):
    __tablename__ = 'tom_nonlocalizedevents_eventlocalization'
    id = sa.Column(sa.Integer, primary_key=True)
    tiles = relationship(lambda: SaSkymapTile)


class SaSkymapTile(Base):
    __tablename__ = 'tom_nonlocalizedevents_skymaptile'
    id = sa.Column(sa.Integer, primary_key=True)
    localization_id = sa.Column(sa.ForeignKey(SaSkymap.id), primary_key=True)
    tile = sa.Column(Tile, primary_key=True, index=True)
    probdensity = sa.Column(sa.Float, nullable=False)
    distance_mean = sa.Column(sa.Float, nullable=False)
    distance_std = sa.Column(sa.Float, nullable=False)


class SaTarget(Base):
    __tablename__ = 'tom_targets_target'
    id = sa.Column(sa.Integer, primary_key=True)
    distance = sa.Column(sa.Float, nullable=True)
    distance_err = sa.Column(sa.Float, nullable=True)


class SaEventCandidate(Base):
    __tablename__ = 'tom_nonlocalizedevents_eventcandidate'
    id = sa.Column(sa.Integer, primary_key=True)
    target_id = sa.Column(sa.ForeignKey(SaTarget.id), primary_key=True)
    healpix = sa.Column(Point, index=True, nullable=False)


def update_all_credible_region_percents_for_candidates(eventlocalization, event_candidate_ids=None):
    ''' This helper function runs through the defined set of discrete credible region probabilities
        and creates or updates the probability of a credible region for each of the event candidates
        that fall within the event localization's credible region of that priority
    '''
    if not event_candidate_ids:
        event_candidate_ids = list(eventlocalization.nonlocalizedevent.candidates.values_list('pk', flat=True))
    for probability in CREDIBLE_REGION_PROBABILITIES:
        update_credible_region_percent_for_candidates(eventlocalization, probability, event_candidate_ids)


def update_credible_region_percent_for_candidates(eventlocalization, prob, event_candidate_ids=None):
    ''' This function creates a credible region linkage with probability prob for each of the event candidate
        ids supplied if they fall within that prob for the event location specified.
    '''
    if not event_candidate_ids:
        event_candidate_ids = list(eventlocalization.nonlocalizedevent.candidates.values_list('pk', flat=True))

    with Session(sa_engine) as session:

        cum_prob = sa.func.sum(
            SaSkymapTile.probdensity * SaSkymapTile.tile.area
        ).over(
            order_by=SaSkymapTile.probdensity.desc()
        ).label(
            'cum_prob'
        )

        subquery = sa.select(
            SaSkymapTile.probdensity,
            cum_prob
        ).filter(
            SaSkymapTile.localization_id == eventlocalization.id
        ).subquery()

        min_probdensity = sa.select(
            sa.func.min(subquery.columns.probdensity)
        ).filter(
            subquery.columns.cum_prob <= prob
        ).scalar_subquery()

        query = sa.select(
            SaEventCandidate.id
        ).filter(
            SaEventCandidate.id.in_(event_candidate_ids),
            SaSkymapTile.localization_id == eventlocalization.id,
            SaSkymapTile.tile.contains(SaEventCandidate.healpix),
            SaSkymapTile.probdensity >= min_probdensity
        )

        results = session.execute(query)

        for sa_event_candidate_id in results:
            CredibleRegion.objects.update_or_create(
                candidate=EventCandidate.objects.get(id=sa_event_candidate_id[0]),
                localization=eventlocalization,
                defaults={
                    'smallest_percent': int(prob * 100.0)
                }
            )

# def point_in_range_django(eventlocalization, prob):
# This code is a beginning attempt to translate the healpix_alchemy query from sql_alchemy to django ORM
# It is non-functional and needs more work to get right.
#     event_candidate_ids = list(eventlocalization.nonlocalizedevent.candidates.values_list('pk', flat=True))

#     SkymapTile.objects.order_by('-probdensity').aggregate(cum_prob=Sum(F('probdensity') * F('tile__area')))

#     cum_prob = SkymapTile.objects.order_by('-probdensity').alias(cum_prob=Sum(F('probdensity') *
#                F('tile__area'))).annotate(cum_prob=F('cum_prob'),
#     )
#     min_probdensity = SkymapTile.objects.order_by('-probdensity').alias(
#         cum_prob=Sum(F('probdensity') * F('tile__area')), min_probdensity=Min(F('probdensity'))).annotate(
#         cum_prob=F('cum_prob'), min_probdensity=F('min_probdensity')
#     ).filter(
#         localization__id=eventlocalization.id,
#         cum_prob__lte=prob,
#         probdensity__gte=min_probdensity,
#         tile__contains=
#         localization__nonlocalizedevent__candidates
#     )
#     candidates = EventCandidate.objects.filter(
#         nonlocalizedevent=eventlocalization.nonlocalizedevent,
#         eventlocalization__tiles__contains(healpix)
#     ).values('id')



import logging
import traceback

from hop.io import Metadata
from hop.models import JSONBlob
from django.conf import settings

from tom_nonlocalizedevents.models import NonLocalizedEvent, EventSequence, ExternalCoincidence
from tom_nonlocalizedevents.healpix_utils import create_localization_for_skymap

logger = logging.getLogger(__name__)


def get_sequence_number(superevent_id: str) -> int:
    """ This returns the sequence number of the next sequence for a superevent_id. This is a hack to get
        around the fact that IGWN GWAlerts no longer tell you the sequence number within them.
    """
    try:
        nle = NonLocalizedEvent.objects.get(event_id=superevent_id)
        return nle.sequences.count() + 1
    except NonLocalizedEvent.DoesNotExist:
        return 1  # The nonlocalizedevent doesnt exist in our system yet, so this must be the first sequence


def handle_igwn_message(message: JSONBlob, metadata: Metadata):
    alert = message.content[0]
    logger.warning(f"Handling igwn alert: {alert}")

    # Only store test alerts if we are configured to do so
    # TODO: consider moving SAVE_TEST_ALERTS from top level of settings
    #  to the ALERT_STREAMS list of stream configuration dictionaries.
    try:
        save_test_alerts = settings.SAVE_TEST_ALERTS
    except AttributeError as err:
        save_test_alerts = True
        logger.warning(f'{err} Using {save_test_alerts} as default value.')
    if alert.get('superevent_id', '').startswith('M') and not save_test_alerts:
        return None, None

    if alert.get('alert_type', '') == 'RETRACTION':
        nonlocalizedevent, nle_created = NonLocalizedEvent.objects.update_or_create(
            event_id=alert['superevent_id'],
            event_type=NonLocalizedEvent.NonLocalizedEventType.GRAVITATIONAL_WAVE,
            defaults={'state': NonLocalizedEvent.NonLocalizedEventState.RETRACTED}
        )
        return nonlocalizedevent, None

    nonlocalizedevent, nle_created = NonLocalizedEvent.objects.get_or_create(
        event_id=alert['superevent_id'],
        event_type=NonLocalizedEvent.NonLocalizedEventType.GRAVITATIONAL_WAVE,
    )
    if nle_created:
        logger.info(f"Ingested a new GW event with id {alert['superevent_id']} from IGWN alertstream")

    # Here we do a bit of pre-processing for IGWN alerts in order to be able to remove the skymap before saving the file
    localization = None
    if alert.get('event'):
        skymap_bytes = alert['event'].pop('skymap')
        if skymap_bytes:
            try:
                localization = create_localization_for_skymap(
                    nonlocalizedevent=nonlocalizedevent, skymap_bytes=skymap_bytes
                )
            except Exception as e:
                localization = None
                logger.error(f'Could not create EventLocalization for event: {alert["superevent_id"]}. Exception: {e}')
                logger.error(traceback.format_exc())

    external_coincidence = None
    if alert.get('external_coinc', {}):
        combined_skymap_bytes = alert['external_coinc'].pop('combined_skymap')
        if combined_skymap_bytes:
            try:
                combined_localization = create_localization_for_skymap(
                    nonlocalizedevent=nonlocalizedevent, skymap_bytes=combined_skymap_bytes, is_combined=True
                )
                external_coincidence, _ = ExternalCoincidence.objects.get_or_create(
                    localization=combined_localization,
                    defaults={'details': alert.get('external_coinc')}
                )
            except Exception as e:
                external_coincidence = None
                logger.error(
                    f'Could not create combined EventLocalization for event: {alert["superevent_id"]}. Exception: {e}'
                )
                logger.error(traceback.format_exc())

    logger.debug(f"Storing igwn alert: {alert}")

    # Now ingest the sequence for that event
    sequence_number = get_sequence_number(alert['superevent_id'])
    event_sequence, es_created = EventSequence.objects.update_or_create(
        nonlocalizedevent=nonlocalizedevent,
        localization=localization,
        external_coincidence=external_coincidence,
        sequence_id=sequence_number,
        details=alert.get('event'),
        defaults={
            'event_subtype': alert.get('alert_type'),
            'ingestor_source': 'hop'
        }
    )
    if es_created and localization is None:
        warning_msg = (
            f'{"Creating" if es_created else "Updating"} EventSequence without EventLocalization:'
            f'{event_sequence} for NonLocalizedEvent: {nonlocalizedevent}'
        )
        logger.warning(warning_msg)

    return nonlocalizedevent, event_sequence



from django.urls import path

from tom_common.api_router import SharedAPIRootRouter  # a singleton DRF Router
from tom_nonlocalizedevents.views import SupereventPkView, SupereventIdView

from . import views

# this mechanism allows ViewSets to be registered with the tom_common Router
# for any of the INSTALLED_APPS (i.e. the routes are added because the APP is
# INSTALLED -- nothing else is required))
router = SharedAPIRootRouter()
router.register(r'nonlocalizedevents', views.NonLocalizedEventViewSet)
router.register(r'eventlocalizations', views.EventLocalizationViewSet)
router.register(r'eventcandidates', views.EventCandidateViewSet)

# app_name provides namespace in {% url %} template tag
# (i.e. {% url 'nonlocalizedevents:detail' <pk> %}
app_name = 'nonlocalizedevents'

urlpatterns = [
    path('', views.NonLocalizedEventListView.as_view(), name='index'),
    path('<int:pk>/', SupereventPkView.as_view(), name='detail'),
    path('<str:event_id>/', SupereventIdView.as_view(), name='event-detail'),
    path('alert/createfrom', views.CreateEventFromHermesAlertView.as_view(), name='create-from-alert')
]



from functools import partial

from django.dispatch import receiver
from django.db.models.signals import post_save
from django.db import transaction

from tom_nonlocalizedevents.models import EventCandidate, EventLocalization
from tom_nonlocalizedevents.healpix_utils import update_all_credible_region_percents_for_candidates

import logging
logger = logging.getLogger(__name__)


@receiver(post_save, sender=EventCandidate)
def cb_post_save_event_candidate(sender, instance, created, **kwargs):
    # Anytime an EventCandidate is created, create and save its probability for different levels of credible region
    # for each localization we have for that event
    if created:
        logger.warning(f"Just created and saved new Event Candidate with healpix {instance.healpix}")
        localizations = instance.nonlocalizedevent.localizations.all()
        for localization in localizations:
            update_all_credible_region_percents_for_candidates(localization, [instance.pk])


@receiver(post_save, sender=EventLocalization)
def cb_post_save_event_localization(sender, instance, created, **kwargs):
    # Anytime a new EventLocalization is created, update the smallest probability credible region for each
    # EventCandidate associated with that nonlocalizedevent for the new localization
    if created:
        transaction.on_commit(
            partial(update_all_credible_region_percents_for_candidates, eventlocalization=instance)
        )



from django.apps import AppConfig


class NonLocalizedEventTypesConfig(AppConfig):
    name = 'tom_nonlocalizedevents'

    def ready(self):
        import tom_nonlocalizedevents.signals.handlers  # noqa
        super().ready()



# Generated by Django 3.1.7 on 2021-10-19 16:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0007_eventcandidate_viable'),
    ]

    operations = [
        migrations.AddField(
            model_name='eventcandidate',
            name='priority',
            field=models.IntegerField(default=1),
        ),
    ]



# Generated by Django 4.2 on 2023-04-28 20:56

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0016_externalcoincidence_alter_eventsequence_options_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eventsequence',
            name='external_coincidence',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='sequences', to='tom_nonlocalizedevents.externalcoincidence'),
        ),
        migrations.AlterField(
            model_name='eventsequence',
            name='ingestor_source',
            field=models.CharField(default='', help_text='The source of this alert - i.e. which stream is it from?', max_length=64),
        ),
    ]



# Generated by Django 3.1.5 on 2021-02-12 19:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Superevent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('superevent_id', models.CharField(max_length=64)),
                ('superevent_url', models.URLField()),
            ],
        ),
    ]



# Generated by Django 3.1.7 on 2021-10-05 20:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0006_auto_20210827_2346'),
    ]

    operations = [
        migrations.AddField(
            model_name='eventcandidate',
            name='viable',
            field=models.BooleanField(default=True),
        ),
    ]



# Generated by Django 4.1.1 on 2022-10-13 20:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0011_eventcandidate_viability_reason_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='nonlocalizedevent',
            name='event_subtype',
            field=models.CharField(default='', help_text='The subtype of the event. Options are type specific, i.e. GW events have initial, preliminary, update types.', max_length=256),
        ),
    ]



# Generated by Django 4.1.2 on 2022-10-25 22:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0013_eventcandidate_healpix_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='CredibleRegion',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('smallest_percent', models.IntegerField(default=100, help_text='Smallest percent credible region this candidate falls into for this localization.')),
            ],
        ),
        migrations.RemoveConstraint(
            model_name='eventcandidate',
            name='Unique Target/NonLocalizedEvent',
        ),
        migrations.AlterField(
            model_name='eventlocalization',
            name='nonlocalizedevent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='localizations', to='tom_nonlocalizedevents.nonlocalizedevent'),
        ),
        migrations.AddConstraint(
            model_name='eventcandidate',
            constraint=models.UniqueConstraint(fields=('target', 'nonlocalizedevent'), name='unique_target_nonlocalizedevent'),
        ),
        migrations.AddField(
            model_name='credibleregion',
            name='candidate',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='credibleregions', to='tom_nonlocalizedevents.eventcandidate'),
        ),
        migrations.AddField(
            model_name='credibleregion',
            name='localization',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='credibleregions', to='tom_nonlocalizedevents.eventlocalization'),
        ),
        migrations.AddConstraint(
            model_name='credibleregion',
            constraint=models.UniqueConstraint(fields=('localization', 'candidate'), name='unique_localization_candidate'),
        ),
    ]



# Generated by Django 4.1.1 on 2022-10-13 00:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0010_rename_more_superevent_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='eventcandidate',
            name='viability_reason',
            field=models.TextField(default='', help_text='Reason why this candidates viability is set the way it is.'),
        ),
        migrations.AlterField(
            model_name='eventcandidate',
            name='viable',
            field=models.BooleanField(default=True, help_text='Whether this event candidate is actively being considered or not'),
        ),
        migrations.AlterField(
            model_name='nonlocalizedevent',
            name='event_id',
            field=models.CharField(help_text='Unique identifer for the event. I.E. the TRIGGER_NUM for a GW event.', max_length=64),
        ),
        migrations.AlterField(
            model_name='nonlocalizedevent',
            name='event_type',
            field=models.CharField(choices=[('GW', 'Gravitational Wave'), ('GRB', 'Gamma-ray Burst'), ('NU', 'Neutrino'), ('UNK', 'Unknown')], default='GW', help_text='The type of NonLocalizedEvent, used for determining how to ingest and display it', max_length=3),
        ),
        migrations.AlterField(
            model_name='nonlocalizedevent',
            name='sequence_id',
            field=models.PositiveIntegerField(default=1, help_text='The version / update number of this event. I.E. the SEQUENCE_NUM for a GW event.'),
        ),
        migrations.AlterField(
            model_name='nonlocalizedevent',
            name='skymap_file_url',
            field=models.URLField(default='', help_text='The URL to a file containing skymap details for the event.'),
        ),
        migrations.AlterUniqueTogether(
            name='nonlocalizedevent',
            unique_together={('event_id', 'sequence_id')},
        ),
    ]



# Generated by Django 4.2 on 2023-04-21 23:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0015_eventsequence_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExternalCoincidence',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('details', models.JSONField(blank=True, help_text='Contains the payload of the external coincidence section of the alert', null=True)),
            ],
        ),
        migrations.AlterModelOptions(
            name='eventsequence',
            options={'ordering': ['sequence_id']},
        ),
        migrations.RemoveConstraint(
            model_name='eventlocalization',
            name='unique_date_per_nonlocalizedevent',
        ),
        migrations.RemoveField(
            model_name='eventlocalization',
            name='skymap_moc_file_url',
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='area_50',
            field=models.FloatField(blank=True, help_text='The 50 percent confidence region area', null=True),
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='area_90',
            field=models.FloatField(blank=True, help_text='The 90 percent confidence region area', null=True),
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='skymap_hash',
            field=models.UUIDField(blank=True, help_text='A UUID from an md5 hash of the raw skymap file contents, used to detect when it has changed', null=True),
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='skymap_url',
            field=models.URLField(default='', help_text='The URL to a file containing skymap file for used to generate this localization.'),
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='skymap_version',
            field=models.PositiveSmallIntegerField(blank=True, help_text='The version of the skymap for this localization', null=True),
        ),
        migrations.AddField(
            model_name='eventsequence',
            name='details',
            field=models.JSONField(blank=True, help_text='Contains the payload of the event section of the alert', null=True),
        ),
        migrations.AddField(
            model_name='eventsequence',
            name='ingestor_source',
            field=models.CharField(default='', help_text='The source of this alert - i.e. which stream is it from?'),
        ),
        migrations.AlterField(
            model_name='eventcandidate',
            name='priority',
            field=models.IntegerField(default=1, help_text='Internal priority of this Event Candidate'),
        ),
        migrations.AddConstraint(
            model_name='eventlocalization',
            constraint=models.UniqueConstraint(fields=('nonlocalizedevent', 'skymap_hash'), name='unique_skymap_per_nonlocalizedevent'),
        ),
        migrations.AddField(
            model_name='externalcoincidence',
            name='localization',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='external_coincidences', to='tom_nonlocalizedevents.eventlocalization'),
        ),
        migrations.AddField(
            model_name='eventsequence',
            name='external_coincidence',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='sequences', to='tom_nonlocalizedevents.externalcoincidence'),
        ),
    ]



# Generated by Django 3.2.6 on 2021-08-27 23:46

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0019_auto_20210811_0018'),
        ('tom_nonlocalizedevents', '0005_auto_20210319_2241'),
    ]

    operations = [
        migrations.CreateModel(
            name='EventCandidate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('superevent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tom_nonlocalizedevents.superevent')),
                ('target', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tom_targets.target')),
            ],
        ),
        migrations.AddConstraint(
            model_name='eventcandidate',
            constraint=models.UniqueConstraint(fields=('target', 'superevent'), name='Unique Target/Superevent'),
        ),
    ]



''' This class defines a message handler for a tom_alertstreams connection to GW events

'''
import logging
import os
import traceback
import requests
from django.conf import settings

from tom_nonlocalizedevents.models import NonLocalizedEvent, EventSequence
from tom_nonlocalizedevents.healpix_utils import create_localization_for_skymap

logger = logging.getLogger(__name__)


EXPECTED_FIELDS = [
    'trigger_num',
    'sequence_num',
    'notice_type',
    'skymap_fits_url'
]


def extract_all_fields(message):
    parsed_fields = {}
    for line in message.splitlines():
        entry = line.split(':', maxsplit=1)
        if len(entry) > 1:
            if entry[0].strip() == 'COMMENTS' and 'comments' in parsed_fields:
                parsed_fields['comments'] += entry[1].lstrip()
            else:
                parsed_fields[entry[0].strip().lower()] = entry[1].strip()
    return parsed_fields


def get_moc_url_from_skymap_fits_url(skymap_fits_url):
    base, filename = os.path.split(skymap_fits_url)
    # Repair broken skymap filenames given in gcn mock alerts right now
    if filename.endswith('.fit'):
        filename = filename + 's'
    # Replace the non-MOC skymap url provided with the MOC version, but keep the ,# on the end
    filename = filename.replace('LALInference.fits.gz', 'LALInference.multiorder.fits')
    filename = filename.replace('bayestar.fits.gz', 'bayestar.multiorder.fits')
    return os.path.join(base, filename)


def handle_message(message):
    # It receives a bytestring message or a Kafka message in the LIGO GW format
    # fields must be extracted from the message text and stored into in the model
    # It returns the nonlocalizedevent and event sequence ingested or None, None.
    if not isinstance(message, bytes):
        bytes_message = message.value()
    else:
        bytes_message = message
    logger.warning(f"Processing message: {bytes_message.decode('utf-8')}")
    fields = extract_all_fields(bytes_message.decode('utf-8'))
    if not all(field in fields.keys() for field in EXPECTED_FIELDS):
        logger.warning(f"Incoming GW message did not have the expected fields, ignoring it: {fields.keys()}")
        return

    if fields and fields['trigger_num'].startswith('M') and not settings.SAVE_TEST_ALERTS:
        return

    if fields:
        nonlocalizedevent, nle_created = NonLocalizedEvent.objects.get_or_create(
            event_id=fields['trigger_num'],
            event_type=NonLocalizedEvent.NonLocalizedEventType.GRAVITATIONAL_WAVE,
        )
        if nle_created:
            logger.info(f"Ingested a new GW event with id {fields['trigger_num']} from alertstream")
        # Next attempt to ingest and build the localization of the event
        skymap_url = get_moc_url_from_skymap_fits_url(fields['skymap_fits_url'])
        try:
            skymap_resp = requests.get(skymap_url)
            skymap_resp.raise_for_status()
            localization = create_localization_for_skymap(
                nonlocalizedevent=nonlocalizedevent,
                skymap_bytes=skymap_resp.content,
                skymap_url=skymap_url
            )
        except Exception as e:
            localization = None
            logger.error(
                f"Failed to retrieve and process localization from skymap file at {skymap_url}. Exception: {e}"
            )
            logger.error(traceback.format_exc())

        # Now ingest the sequence for that event
        event_sequence, es_created = EventSequence.objects.update_or_create(
            nonlocalizedevent=nonlocalizedevent,
            localization=localization,
            sequence_id=fields['sequence_num'],
            defaults={
                'event_subtype': fields['notice_type'],
                'details': fields,
                'ingestor_source': 'gcn'
            }
        )
        if es_created and localization is None:
            warning_msg = (f'{"Creating" if es_created else "Updating"} EventSequence without EventLocalization:'
                           f'{event_sequence} for NonLocalizedEvent: {nonlocalizedevent}')
            logger.warning(warning_msg)

        return nonlocalizedevent, event_sequence
    return None, None


def handle_retraction(message):
    # It receives a bytestring message or a Kafka message in the LIGO GW format
    # For a retraction message, we just set the events status to retracted if it exists.
    if not isinstance(message, bytes):
        bytes_message = message.value()
    else:
        bytes_message = message
    # Just need the event_id from the retraction messages
    fields = extract_all_fields(bytes_message.decode('utf-8'))
    if 'trigger_num' not in fields:
        logger.warning("Retraction notice missing 'trigger_num' field, ignoring.")
        return
    # Then set the state to 'RETRACTED' for the event matching that id
    try:
        retracted_event = NonLocalizedEvent.objects.get(event_id=fields['trigger_num'])
        retracted_event.state = NonLocalizedEvent.NonLocalizedEventState.RETRACTED
        retracted_event.save()
        return retracted_event
    except NonLocalizedEvent.DoesNotExist:
        logger.warning((f"Got a Retraction notice for event id {fields['trigger_num']}"
                        f"which does not exist in the database"))
    return None



from django.db import models
from django.contrib.postgres.fields import BigIntegerRangeField
from django.contrib.postgres.indexes import SpGistIndex
from django.conf import settings

from tom_targets.models import Target

import numpy as np
from psycopg2.extensions import register_adapter, AsIs
from healpix_alchemy.constants import HPX
from astropy.coordinates import SkyCoord
from urllib.parse import urljoin
import logging


def adapt_numpy_float64(np_float64):
    return AsIs(np_float64)


def adapt_numpy_int64(np_int64):
    return AsIs(np_int64)


register_adapter(np.float64, adapt_numpy_float64)
register_adapter(np.int64, adapt_numpy_int64)


logger = logging.getLogger(__name__)


class NonLocalizedEvent(models.Model):
    """Represents a NonLocalizedEvent being followed-up upon by this TOM.

    A NonLocalizedEvent is distinguished from a Target in that it is localized to a region of the sky
    (vs. a specific RA,DEC). The potential Targets in the localization region must be identified,
    prioritized, and categorized (retired, of-interest, etc) for follow-up EM observations

    For the moment, this is rather GraceDB (GW) specific, but sh/could be generalized to work
    with gamma-ray burst and neutrino events.
    """

    class NonLocalizedEventType(models.TextChoices):
        GRAVITATIONAL_WAVE = 'GW', 'Gravitational Wave'
        GAMMA_RAY_BURST = 'GRB', 'Gamma-ray Burst'
        NEUTRINO = 'NU', 'Neutrino'
        UNKNOWN = 'UNK', 'Unknown'

    class NonLocalizedEventState(models.TextChoices):
        ACTIVE = 'ACTIVE'
        INACTIVE = 'INACTIVE'
        RETRACTED = 'RETRACTED'

    state = models.CharField(
        max_length=16,
        choices=NonLocalizedEventState.choices,
        default=NonLocalizedEventState.ACTIVE,
        help_text='The current state of this NonLocalizedEvent'
    )

    event_type = models.CharField(
        max_length=3,
        choices=NonLocalizedEventType.choices,
        default=NonLocalizedEventType.GRAVITATIONAL_WAVE,
        help_text='The type of NonLocalizedEvent, used for determining how to ingest and display it'
    )
    # TODO: ask Curtis/Rachel/Andy about generalized use cases.
    event_id = models.CharField(
        max_length=64,
        help_text='Unique identifer for the event. I.E. the TRIGGER_NUM for a GW event.'
    )

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['event_id'], name='unique_event_id')
        ]

    @property
    def gracedb_url(self):
        """Construct and return the GraceDB URL for this nonlocalizedevent from the event_id Field.

        for example, https://gracedb.ligo.org/superevents/S200316bj/
        """
        # TODO: add check that superevent_type is GRAVITATIONAL_WAVE
        return f"https://gracedb.ligo.org/superevents/{self.event_id}/"

    @property
    def treasuremap_url(self):
        """Construct and return the Treasure Map (treasuremap.space) URL for this nonlocalizedevent
        from the event_id Field.

        for example: http://treasuremap.space/alerts?graceids=S200219ac
        """
        # TODO: add check that superevent_type is GRAVITATIONAL_WAVE
        return f"http://treasuremap.space/alerts?graceids={self.event_id}"

    @property
    def hermes_url(self):
        """Construct and return the hermes details page URL for this nonlocalizedevent

        for example: http://hermes.lco.global/nonlocalizedevent/S200316bj/
        """
        return urljoin(settings.HERMES_API_URL, f'/nonlocalizedevent/{self.event_id}/')

    def __str__(self):
        return f"{self.event_id}"


class EventLocalization(models.Model):
    """Represents a region of the sky in which a nonlocalizedevent may have taken place.
    """
    nonlocalizedevent = models.ForeignKey(NonLocalizedEvent, related_name='localizations', on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    date = models.DateTimeField(
        help_text='The datestamp of this localizations creation.'
    )
    skymap_url = models.URLField(
        default='',
        help_text='The URL to a file containing skymap file for used to generate this localization.'
    )
    skymap_version = models.PositiveSmallIntegerField(
        null=True, blank=True,
        help_text='The version of the skymap for this localization'
    )
    skymap_hash = models.UUIDField(
        null=True, blank=True,
        help_text='A UUID from an md5 hash of the raw skymap file contents, used to detect when it has changed'
    )
    distance_mean = models.FloatField(
        default=0,
        help_text='The posterior mean distance in Mpc.'
    )
    distance_std = models.FloatField(
        default=0,
        help_text='The posterior standard deviation of the distance in Mpc.'
    )
    area_50 = models.FloatField(
        null=True, blank=True,
        help_text='The 50 percent confidence region area'
    )
    area_90 = models.FloatField(
        null=True, blank=True,
        help_text='The 90 percent confidence region area'
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['nonlocalizedevent', 'skymap_hash'],
                name='unique_skymap_per_nonlocalizedevent'
            )
        ]


class ExternalCoincidence(models.Model):
    details = models.JSONField(
        null=True, blank=True,
        help_text='Contains the payload of the external coincidence section of the alert'
    )
    localization = models.ForeignKey(
        EventLocalization, related_name='external_coincidences',
        null=True, on_delete=models.SET_NULL
    )


class EventSequence(models.Model):
    nonlocalizedevent = models.ForeignKey(NonLocalizedEvent, related_name='sequences', on_delete=models.CASCADE)
    localization = models.ForeignKey(EventLocalization, related_name='sequences', null=True, on_delete=models.SET_NULL)
    external_coincidence = models.ForeignKey(
        ExternalCoincidence, related_name='sequences',
        null=True, blank=True, on_delete=models.SET_NULL
    )
    details = models.JSONField(
        null=True, blank=True,
        help_text='Contains the payload of the event section of the alert'
    )
    sequence_id = models.PositiveIntegerField(
        default=1,
        help_text='The version / update number of this event. I.E. the SEQUENCE_NUM for a GW event.'
    )
    event_subtype = models.CharField(
        max_length=256,
        default='',
        help_text='The subtype of the event. Options are type specific, i.e. GW events have initial, '
                  'preliminary, update types.'
    )
    ingestor_source = models.CharField(
        default='', max_length=64,
        help_text='The source of this alert - i.e. which stream is it from?'
    )
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ["sequence_id"]
        constraints = [
            models.UniqueConstraint(
                fields=['nonlocalizedevent', 'sequence_id'],
                name='unique_sequence_per_nonlocalizedevent'
            )
        ]


class EventCandidate(models.Model):
    target = models.ForeignKey(Target, on_delete=models.CASCADE)
    nonlocalizedevent = models.ForeignKey(NonLocalizedEvent, related_name='candidates', on_delete=models.CASCADE)

    viable = models.BooleanField(
        default=True,
        help_text='Whether this event candidate is actively being considered or not'
    )
    viability_reason = models.TextField(
        default='',
        help_text='Reason why this candidates viability is set the way it is.'
    )
    priority = models.IntegerField(
        default=1,
        help_text='Internal priority of this Event Candidate'
    )
    healpix = models.BigIntegerField(
        default=-1,
        help_text='Stores the healpix index for this candidates target, used for MOC lookups'
    )
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    def save(self, *args, **kwargs):
        if self.target:
            self.healpix = HPX.skycoord_to_healpix(SkyCoord(self.target.ra, self.target.dec, unit='deg'))
        super().save(*args, **kwargs)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['target', 'nonlocalizedevent'], name='unique_target_nonlocalizedevent')
        ]

    def __str__(self):
        return f'EventCandidate({self.id}) NonLocalizedEvent: {self.nonlocalizedevent} Target: {self.target}'


class CredibleRegion(models.Model):
    localization = models.ForeignKey(EventLocalization, related_name='credibleregions', on_delete=models.CASCADE)
    candidate = models.ForeignKey(EventCandidate, related_name='credibleregions', on_delete=models.CASCADE)

    smallest_percent = models.IntegerField(
        default=100,
        help_text='Smallest percent credible region this candidate falls into for this localization.'
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['localization', 'candidate'], name='unique_localization_candidate')
        ]


class SkymapTile(models.Model):
    """ A healpix_alchemy style tile linked to an event localization
    """
    localization = models.ForeignKey(EventLocalization, related_name='tiles', on_delete=models.CASCADE)
    tile = BigIntegerRangeField(db_index=True)
    probdensity = models.FloatField(null=False)
    distance_mean = models.FloatField(
        default=0
    )
    distance_std = models.FloatField(
        default=0
    )

    class Meta:
        indexes = [SpGistIndex(fields=('tile',))]



# Generated by Django 4.1.2 on 2022-10-27 20:37

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0014_credibleregion_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='EventSequence',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sequence_id', models.PositiveIntegerField(default=1, help_text='The version / update number of this event. I.E. the SEQUENCE_NUM for a GW event.')),
                ('event_subtype', models.CharField(default='', help_text='The subtype of the event. Options are type specific, i.e. GW events have initial, preliminary, update types.', max_length=256)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='nonlocalizedevent',
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name='eventlocalization',
            name='sequence_id',
        ),
        migrations.AddField(
            model_name='eventcandidate',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='eventcandidate',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='date',
            field=models.DateTimeField(default=django.utils.timezone.now, help_text='The datestamp of this localizations creation.'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='distance_mean',
            field=models.FloatField(default=0, help_text='The posterior mean distance in Mpc.'),
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='distance_std',
            field=models.FloatField(default=0, help_text='The posterior standard deviation of the distance in Mpc.'),
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='skymap_moc_file_url',
            field=models.URLField(default='', help_text='The URL to a file containing skymap MOC file for the event sequence. Used to generate localization.'),
        ),
        migrations.AddField(
            model_name='skymaptile',
            name='distance_mean',
            field=models.FloatField(default=0),
        ),
        migrations.AddField(
            model_name='skymaptile',
            name='distance_std',
            field=models.FloatField(default=0),
        ),
        migrations.AddConstraint(
            model_name='eventlocalization',
            constraint=models.UniqueConstraint(fields=('nonlocalizedevent', 'date'), name='unique_date_per_nonlocalizedevent'),
        ),
        migrations.AddConstraint(
            model_name='nonlocalizedevent',
            constraint=models.UniqueConstraint(fields=('event_id',), name='unique_event_id'),
        ),
        migrations.AddField(
            model_name='eventsequence',
            name='localization',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='sequences', to='tom_nonlocalizedevents.eventlocalization'),
        ),
        migrations.AddField(
            model_name='eventsequence',
            name='nonlocalizedevent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sequences', to='tom_nonlocalizedevents.nonlocalizedevent'),
        ),
        migrations.RemoveField(
            model_name='nonlocalizedevent',
            name='event_subtype',
        ),
        migrations.RemoveField(
            model_name='nonlocalizedevent',
            name='sequence_id',
        ),
        migrations.RemoveField(
            model_name='nonlocalizedevent',
            name='skymap_file_url',
        ),
        migrations.AddConstraint(
            model_name='eventsequence',
            constraint=models.UniqueConstraint(fields=('nonlocalizedevent', 'sequence_id'), name='unique_sequence_per_nonlocalizedevent'),
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-19 22:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0004_superevent_superevent_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='superevent',
            name='superevent_type',
            field=models.CharField(choices=[('GW', 'Gravitational Wave'), ('GRB', 'Gamma-ray Burst'), ('NU', 'Neutrino'), ('UNK', 'Unknown')], default='GW', max_length=3),
        ),
    ]



from django.contrib import admin

from .models import (EventCandidate, EventLocalization,
                     NonLocalizedEvent, EventSequence,
                     ExternalCoincidence, CredibleRegion)

# Register your models here.
admin.site.register(CredibleRegion)
admin.site.register(EventCandidate)
admin.site.register(EventLocalization)
admin.site.register(EventSequence)
admin.site.register(NonLocalizedEvent)
admin.site.register(ExternalCoincidence)



import factory
# from datetime import datetime
from django.db.models import signals
from tom_nonlocalizedevents.models import NonLocalizedEvent, EventLocalization, EventSequence


class NonLocalizedEventFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = NonLocalizedEvent

    event_id = factory.Faker('pystr')


class EventSequenceFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EventSequence


@factory.django.mute_signals(signals.post_save)
class EventLocalizationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EventLocalization

    skymap_url = factory.Faker('pystr')
    date = factory.Faker('date_time')



# Generated by Django 4.1.1 on 2022-10-13 00:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0008_eventcandidate_priority'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='Superevent',
            new_name='NonLocalizedEvent',
        ),
    ]



from django.contrib.auth.models import User
from django.urls import reverse
from rest_framework.test import APITestCase

from tom_nonlocalizedevents.tests.factories import (NonLocalizedEventFactory, EventLocalizationFactory,
                                                    EventSequenceFactory)


class NonLocalizedEventAPITestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username='test_user')
        self.superevent1 = NonLocalizedEventFactory.create(event_id='superevent1')
        self.superevent2 = NonLocalizedEventFactory.create(event_id='superevent2')
        self.eventlocalization1 = EventLocalizationFactory.create(nonlocalizedevent=self.superevent1)
        self.eventlocalization2 = EventLocalizationFactory.create(nonlocalizedevent=self.superevent2)
        self.sequence11 = EventSequenceFactory.create(
            nonlocalizedevent=self.superevent1, localization=self.eventlocalization1
        )
        self.sequence21 = EventSequenceFactory.create(
            nonlocalizedevent=self.superevent2, localization=self.eventlocalization2
        )
        self.client.force_login(self.user)


class TestNonLocalizedEventViewSet(NonLocalizedEventAPITestCase):

    def test_nonlocalizedevent_list_api(self):
        """Test NonLocalizedEvent API list endpoint."""
        response = self.client.get(reverse('api:nonlocalizedevent-list'))

        self.assertEqual(response.json()['count'], 2)
        self.assertContains(response, f'"event_id":"{self.superevent1.event_id}"')
        self.assertContains(response, f'"skymap_url":"{self.eventlocalization1.skymap_url}"')
        self.assertContains(response, f'"event_id":"{self.superevent2.event_id}"')
        self.assertContains(response, f'"skymap_url":"{self.eventlocalization2.skymap_url}"')

    def test_nonlocalizedevent_index_view(self):
        response = self.client.get(reverse('nonlocalizedevents:index'))

        self.assertContains(response, self.superevent1.event_id)
        self.assertContains(response, self.superevent2.event_id)
        self.assertContains(response, reverse('nonlocalizedevents:detail', args=(self.superevent1.pk,)))
        self.assertContains(response, reverse('nonlocalizedevents:detail', args=(self.superevent2.pk,)))

    def test_superevent_detail_view(self):
        response = self.client.get(reverse('nonlocalizedevents:detail', args=(self.superevent1.pk,)))

        self.assertContains(response, self.superevent1.event_id)
        self.assertContains(response, "superevent-sequences")
        self.assertContains(response, "vue")


class TestEventLocalizationViewSet(NonLocalizedEventAPITestCase):
    def test_eventlocalization_list(self):
        """Test EventLocalization API list endpoint."""
        response = self.client.get(reverse('api:eventlocalization-list'))

        self.assertEqual(response.json()['count'], 2)



# Generated by Django 4.1.1 on 2022-10-13 00:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0009_rename_superevent_nonlocalizedevent'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='eventcandidate',
            name='Unique Target/Superevent',
        ),
        migrations.RenameField(
            model_name='eventcandidate',
            old_name='superevent',
            new_name='nonlocalizedevent',
        ),
        migrations.RenameField(
            model_name='nonlocalizedevent',
            old_name='superevent_id',
            new_name='event_id',
        ),
        migrations.RenameField(
            model_name='nonlocalizedevent',
            old_name='superevent_type',
            new_name='event_type',
        ),
        migrations.RenameField(
            model_name='nonlocalizedevent',
            old_name='superevent_url',
            new_name='skymap_file_url',
        ),
        migrations.AddField(
            model_name='nonlocalizedevent',
            name='sequence_id',
            field=models.PositiveIntegerField(default=1),
        ),
        migrations.AddConstraint(
            model_name='eventcandidate',
            constraint=models.UniqueConstraint(fields=('target', 'nonlocalizedevent'), name='Unique Target/NonLocalizedEvent'),
        ),
    ]



import json
import logging

from django.contrib import messages
from django.core.cache import cache
from django.http import Http404
from django.shortcuts import redirect
from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic import ListView, TemplateView
from django.views.generic.base import View
from django.urls import reverse
from django.conf import settings

from rest_framework import permissions, viewsets
from django_filters.rest_framework import DjangoFilterBackend

from tom_nonlocalizedevents.ingestion import ingest_sequence_from_hermes_message
from tom_nonlocalizedevents.models import EventCandidate, EventLocalization, NonLocalizedEvent
from tom_nonlocalizedevents.serializers import (EventCandidateSerializer, EventLocalizationSerializer,
                                                NonLocalizedEventSerializer)


logger = logging.getLogger(__name__)


class NonLocalizedEventListView(LoginRequiredMixin, ListView):
    """
    Unadorned Django ListView subclass for NonLocalizedEvent model.
    """
    model = NonLocalizedEvent
    template_name = 'tom_nonlocalizedevents/index.html'

    def get_queryset(self):
        # '-created' is most recent first
        qs = NonLocalizedEvent.objects.order_by('-created')
        return qs


# from the tom_alerts query_result.html
class CreateEventFromHermesAlertView(View):
    """
    Creates the models.NonLocalizedEvent instance and redirect to NonLocalizedEventDetailView
    """

    def post(self, request, *args, **kwargs):
        """
        """
        # the request.POST is a QueryDict object;
        query_id = self.request.POST['query_id']

        # events is a list[str] of NonLocalizedEvent event_id's: (e.g. 'MS230417a')
        # (i.e the selected events from the query result form)
        events = request.POST.getlist('events', [])

        # if the user didn't select an alert; warn and re-direct back
        if not events:
            messages.warning(request, 'Please select at least one Event to create.')
            reverse_url: str = reverse('tom_alerts:run', kwargs={'pk': query_id})
            return redirect(reverse_url)

        # Create NonLocalizedEvents for each of the selected events.
        for event_id in events:
            logger.debug(f'Creating event {event_id}...')
            # extract the Hermes event from the cache
            # (it was cached by tom_alerts.views.py::RunQueryView)
            cached_event = json.loads(cache.get(f'alert_{event_id}'))

            # early return: alert not in cache
            if not cached_event:
                messages.error(request, 'Could not createn event(s). Try re-running the query to refresh the cache.')
                return redirect(reverse('tom_alerts:run', kwargs={'pk': query_id}))

            # the NonLocalizedEvent is created by handling all the messages from
            # the event sequence as if they were ingested
            for sequenced_message in cached_event['sequences']:
                logger.debug(f"Creating sequence from HermesBroker: {sequenced_message}")
                ingest_sequence_from_hermes_message(sequenced_message)

        return redirect(reverse('nonlocalizedevents:index'))


#
# Django Rest Framework Views
#

class NonLocalizedEventViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint that allows NonLocalizedEvents to be viewed or edited.
    """
    queryset = NonLocalizedEvent.objects.all()
    serializer_class = NonLocalizedEventSerializer
    permission_classes = [permissions.IsAuthenticated]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['event_id', 'event_type']


class EventCandidateViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint for EventCandidate model.

    Implementation has changes for bulk_create and update/PATCH EventCandidate instances.
    """
    queryset = EventCandidate.objects.all()
    serializer_class = EventCandidateSerializer
    permission_classes = [permissions.IsAuthenticated]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['nonlocalizedevent', 'viable', 'priority']

    def get_serializer(self, *args, **kwargs):
        # In order to ensure the list_serializer_class is used for bulk_create, we check that the POST data is a list
        # and add `many = True` to the kwargs
        if isinstance(kwargs.get('data', {}), list):
            kwargs['many'] = True

        return super().get_serializer(*args, **kwargs)

    def update(self, request, *args, **kwargs):
        """Provide support for the PATCH HTTP verb to update individual model fields.

        An example request might look like:

            PATCH http://localhost:8000/api/eventcandidates/18/

        with a Request Body of:

            {
                "viability": false
            }

        """
        kwargs['partial'] = True
        return super().update(request, *args, **kwargs)


class EventLocalizationViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint that allows EventLocalizations to be viewed or edited.
    """
    queryset = EventLocalization.objects.all()
    serializer_class = EventLocalizationSerializer
    permission_classes = [permissions.IsAuthenticated]


class SupereventPkView(LoginRequiredMixin, TemplateView):
    template_name = 'tom_nonlocalizedevents/superevent_vue_app.html'

    def get_context_data(self, **kwargs: dict) -> dict:
        context = super().get_context_data(**kwargs)
        try:
            superevent = NonLocalizedEvent.objects.get(pk=kwargs['pk'])
            data = NonLocalizedEventSerializer(instance=superevent).data
            context['superevent_data'] = json.dumps(data)
            context['tom_api_url'] = settings.TOM_API_URL
            context['hermes_api_url'] = settings.HERMES_API_URL
            return context
        except NonLocalizedEvent.DoesNotExist:
            raise Http404


class SupereventIdView(LoginRequiredMixin, TemplateView):
    template_name = 'tom_nonlocalizedevents/superevent_vue_app.html'

    def get_context_data(self, **kwargs: dict) -> dict:
        context = super().get_context_data(**kwargs)
        try:
            superevent = NonLocalizedEvent.objects.get(event_id=kwargs['event_id'])
            data = NonLocalizedEventSerializer(instance=superevent).data
            context['superevent_data'] = json.dumps(data)
            context['tom_api_url'] = settings.TOM_API_URL
            context['hermes_api_url'] = settings.HERMES_API_URL
            return context
        except NonLocalizedEvent.DoesNotExist:
            raise Http404



# Generated by Django 4.1.2 on 2022-10-22 00:11

import django.contrib.postgres.fields.ranges
import django.contrib.postgres.indexes
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0012_nonlocalizedevent_event_subtype'),
    ]

    operations = [
        migrations.AddField(
            model_name='eventcandidate',
            name='healpix',
            field=models.BigIntegerField(default=-1, help_text='Stores the healpix index for this candidates target, used for MOC lookups'),
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='nonlocalizedevent',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='localization', to='tom_nonlocalizedevents.nonlocalizedevent'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='eventlocalization',
            name='sequence_id',
            field=models.PositiveIntegerField(default=1),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='nonlocalizedevent',
            name='state',
            field=models.CharField(choices=[('ACTIVE', 'Active'), ('INACTIVE', 'Inactive'), ('RETRACTED', 'Retracted')], default='ACTIVE', help_text='The current state of this NonLocalizedEvent', max_length=16),
        ),
        migrations.AlterField(
            model_name='eventcandidate',
            name='nonlocalizedevent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='candidates', to='tom_nonlocalizedevents.nonlocalizedevent'),
        ),
        migrations.CreateModel(
            name='SkymapTile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('tile', django.contrib.postgres.fields.ranges.BigIntegerRangeField(db_index=True)),
                ('probdensity', models.FloatField()),
                ('localization', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tiles', to='tom_nonlocalizedevents.eventlocalization')),
            ],
        ),
        migrations.AddIndex(
            model_name='skymaptile',
            index=django.contrib.postgres.indexes.SpGistIndex(fields=['tile'], name='tom_nonloca_tile_3cc24d_spgist'),
        ),
    ]



from rest_framework import serializers
from tom_targets.models import Target
from tom_targets.serializers import TargetSerializer

from tom_nonlocalizedevents.models import (CredibleRegion, EventCandidate, EventLocalization,
                                           EventSequence, NonLocalizedEvent, ExternalCoincidence)

# from healpix_alchemy.constants import PIXEL_AREA, HPX
# from astropy.coordinates import SkyCoord
import logging

logger = logging.getLogger(__name__)


class CredibleRegionSerializer(serializers.ModelSerializer):

    class Meta:
        model = CredibleRegion
        fields = ['smallest_percent', 'localization', '']


class EventCandidateSerializer(serializers.ModelSerializer):
    """
    Serializer class for the ``EventCandidate``. ``PrimaryKeyRelatedField``s are used in order to allow creating an
    ``EventCandidate`` with just a primary key, and ``to_representation`` is then overridden for proper display values.
    See: https://www.django-rest-framework.org/api-guide/relations/#custom-relational-fields
    """
    nonlocalizedevent = serializers.PrimaryKeyRelatedField(queryset=NonLocalizedEvent.objects.all())
    target = serializers.PrimaryKeyRelatedField(queryset=Target.objects.all(), required=False)
    target_fields = serializers.DictField(required=False, write_only=True)
    credible_regions = serializers.SerializerMethodField()
    UPDATE_KEYS = ['viable', 'viability_reason', 'priority']

    class Meta:
        model = EventCandidate
        fields = '__all__'
        # list_serializer_class = BulkCreateEventCandidateListSerializer

    def validate(self, data):
        if self.context.get('request').method == 'PATCH':
            # Patch requests on candidates should just be used to change viable boolean and reason and priority
            if not any([key in data for key in self.UPDATE_KEYS]):
                raise serializers.ValidationError(
                    f"PATCH update must contain at least one of {self.UPDATE_KEYS.join(', ')} to update"
                )
        else:
            if 'target' not in data and 'target_fields' not in data:
                raise serializers.ValidationError(
                    "Must specify either target or target_fields to create an EventCandidate"
                )
            if 'target_fields' in data and data['target_fields'].get('name'):
                try:
                    target = Target.objects.get(name=data['target_fields']['name'])
                    data['target'] = target
                    if EventCandidate.objects.filter(
                        target=target, nonlocalizedevent=data['nonlocalizedevent']
                    ).exists():
                        raise serializers.ValidationError(
                            f"Event Candidate already exists for target {target.name} "
                            f"and nonlocalizedevent {data['nonlocalizedevent'].event_id}")
                    del data['target_fields']
                except Target.DoesNotExist:
                    target_serializer = TargetSerializer(data=data['target_fields'])
                    target_serializer.is_valid()
                    data['target_fields'] = target_serializer.validated_data
        return super().validate(data)

    def update(self, instance, validated_data):
        instance.viable = validated_data.get('viable', instance.viable)
        instance.viability_reason = validated_data.get('viability_reason', instance.viability_reason)
        instance.priority = validated_data.get('priority', instance.priority)
        instance.save()
        return instance

    def create(self, validated_data):
        if 'target_fields' in validated_data:
            try:
                target = Target.objects.get(name=validated_data['target_fields']['name'])
            except Target.DoesNotExist:
                target_serializer = TargetSerializer(data=validated_data['target_fields'])
                if target_serializer.is_valid():
                    target = target_serializer.save()
                    logger.info(f"Created target {target.id} with name {target.name}")
            validated_data['target'] = target
            del validated_data['target_fields']
        return super().create(validated_data)

    def get_credible_regions(self, instance):
        localization_id_to_credible_region_percent = {}
        credibleregions = instance.credibleregions.all()
        for cr in credibleregions:
            localization_id_to_credible_region_percent[cr.localization.id] = cr.smallest_percent
        return localization_id_to_credible_region_percent

    def to_representation(self, instance):
        representation = super().to_representation(instance)
        # TODO: this is a little unorthodox. it's convenient for now to handle the data from
        # the ForiengKey objects, but that should be handled separately by their own serializers.
        # (and to_representation could be left undisturbed).
        representation['target'] = TargetSerializer(Target.objects.get(pk=representation['target'])).data
        representation['nonlocalizedevent'] = NonLocalizedEvent.objects.get(
            pk=representation['nonlocalizedevent']).event_id
        return representation


class EventLocalizationSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = EventLocalization
        fields = ['id', 'date', 'skymap_url', 'skymap_version', 'area_50', 'area_90', 'distance_mean', 'distance_std']


class ExternalCoincidence(serializers.HyperlinkedModelSerializer):
    localization = EventLocalizationSerializer(read_only=True)

    class Meta:
        model = ExternalCoincidence
        fields = ['details', 'localization']


class EventSequenceSerializer(serializers.HyperlinkedModelSerializer):
    localization = EventLocalizationSerializer(read_only=True)
    external_coincidence = ExternalCoincidence(read_only=True)

    class Meta:
        model = EventSequence
        fields = ['id', 'created', 'modified', 'sequence_id', 'event_subtype',
                  'details', 'ingestor_source', 'localization', 'external_coincidence']


class NonLocalizedEventSerializer(serializers.HyperlinkedModelSerializer):
    candidates = serializers.SerializerMethodField()
    sequences = EventSequenceSerializer(many=True)

    class Meta:
        model = NonLocalizedEvent
        fields = ['id', 'event_id', 'state', 'candidates', 'created', 'modified', 'sequences']

    def get_candidates(self, instance):
        alerts = instance.candidates.all()
        # This returns the nonlocalied event identifier, which means it's duplicated in the response.
        # The NonLocalizedEventSerializer should therefore use its own custom EventCandidateSerializer
        # rather than the one defined above.
        return EventCandidateSerializer(alerts, many=True).data



# Generated by Django 3.2b1 on 2021-03-09 19:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_nonlocalizedevents', '0003_auto_20210225_0034'),
    ]

    operations = [
        migrations.AddField(
            model_name='superevent',
            name='superevent_type',
            field=models.CharField(default='', max_length=50),
            preserve_default=False,
        ),
    ]



import logging
import requests
import traceback

from tom_nonlocalizedevents.healpix_utils import create_localization_for_skymap
from tom_nonlocalizedevents.models import NonLocalizedEvent, EventSequence, ExternalCoincidence


logger = logging.getLogger(__name__)


def ingest_sequence_from_hermes_message(message):
    event_id = message.get('message', {}).get('data', {}).get('superevent_id', None)
    if not event_id:
        logger.warn(f"Got a malformed HermesBroker message, cannot created nonlocalizedevent. Message = {message}")
        return

    data = message.get('message', {}).get('data', {})

    if data.get('alert_type', '') == 'RETRACTION':
        NonLocalizedEvent.objects.update_or_create(
            event_id=event_id,
            event_type=NonLocalizedEvent.NonLocalizedEventType.GRAVITATIONAL_WAVE,
            defaults={'state': NonLocalizedEvent.NonLocalizedEventState.RETRACTED}
        )
        return

    nonlocalizedevent, nle_created = NonLocalizedEvent.objects.get_or_create(
        event_id=event_id,
        event_type=NonLocalizedEvent.NonLocalizedEventType.GRAVITATIONAL_WAVE,
    )
    if nle_created:
        logger.info(f"Ingested a new GW event with id {event_id} from HermesBroker")

    localization = None
    skymap_url = data.get('urls', {}).get('skymap')
    if skymap_url:
        try:
            skymap_resp = requests.get(skymap_url)
            skymap_resp.raise_for_status()
            localization = create_localization_for_skymap(
                nonlocalizedevent=nonlocalizedevent,
                skymap_bytes=skymap_resp.content,
                skymap_url=skymap_url
            )
        except Exception as e:
            localization = None
            logger.error(
                f"Failed to retrieve and process localization from skymap file at {skymap_url}. Exception: {e}"
            )
            logger.error(traceback.format_exc())

    combined_skymap_url = data.get('urls', {}).get('combined_skymap')
    external_coincidence = None
    if combined_skymap_url:
        try:
            combined_skymap_resp = requests.get(combined_skymap_url)
            combined_skymap_resp.raise_for_status()
            combined_localization = create_localization_for_skymap(
                nonlocalizedevent=nonlocalizedevent,
                skymap_bytes=combined_skymap_resp.content,
                skymap_url=combined_skymap_url,
                is_combined=True
            )
            external_coincidence, _ = ExternalCoincidence.objects.get_or_create(
                localization=combined_localization,
                defaults={'details': data.get('external_coinc')}
            )
        except Exception as e:
            combined_localization = None
            logger.error((
                f"Failed to retrieve and process combined localization from combined skymap file at"
                f"{combined_skymap_url}. Exception: {e}"
            ))
            logger.error(traceback.format_exc())

    # Now ingest the sequence for that event
    sequence_number = data.get('sequence_num')
    event_sequence, es_created = EventSequence.objects.update_or_create(
        nonlocalizedevent=nonlocalizedevent,
        localization=localization,
        external_coincidence=external_coincidence,
        sequence_id=sequence_number,
        details=data.get('event'),
        defaults={
            'event_subtype': data.get('alert_type'),
            'ingestor_source': 'hop'
        }
    )
    if es_created and localization is None:
        warning_msg = (
            f'{"Creating" if es_created else "Updating"} EventSequence without EventLocalization:'
            f'{event_sequence} for NonLocalizedEvent: {nonlocalizedevent}'
        )
        logger.warning(warning_msg)
